/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TAppDecTop.cpp
    \brief    Decoder application class
*/

#include <list>
#include <vector>
#include <stdio.h>
#include <fcntl.h>
#include <assert.h>

#include "TAppDecTop.h"
#include "TLibDecoder/AnnexBread.h"
#include "TLibDecoder/NALread.h"

//! \ingroup TAppDecoder
//! \{

// ====================================================================================================================
// Constructor / destructor / initialization / destroy
// ====================================================================================================================

TAppDecTop::TAppDecTop()
{
  ::memset (m_abDecFlag, 0, sizeof (m_abDecFlag));
#ifdef HEVC_SVC
  for(Int i=0; i<MAX_LAYERS; i++)
    m_iPOCLastDisplay[i]  = -MAX_INT;
#else
  m_iPOCLastDisplay  = -MAX_INT;
#endif
}

Void TAppDecTop::create()
{
}

Void TAppDecTop::destroy()
{
#ifndef HEVC_SVC
  if (m_pchBitstreamFile)
  {
    free (m_pchBitstreamFile);
    m_pchBitstreamFile = NULL;
  }
  if (m_pchReconFile)
  {
    free (m_pchReconFile);
    m_pchReconFile = NULL;
  }
#endif
}

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

/**
 - create internal class
 - initialize internal class
 - until the end of the bitstream, call decoding function in TDecTop class
 - delete allocated buffers
 - destroy internal class
 .
 */
#ifdef HEVC_SVC
Void TAppDecTop::decode()
{
  UInt                uiPOC;
  TComList<TComPic*>* pcListPic[MAX_LAYERS] = {NULL};

#if AVC_BL_ENABLE
  UInt      numNalRead = 0;
#endif

  ifstream bitstreamFile(m_pchBitstreamFile, ifstream::in | ifstream::binary);
  if (!bitstreamFile)
  {
    fprintf(stderr, "\nfailed to open bitstream file `%s' for reading\n", m_pchBitstreamFile);
    exit(EXIT_FAILURE);
  }

  InputByteStream bytestream(bitstreamFile);

#if AVC_BL_ENABLE
  m_cTVideoIOYuvBLReconFile.open( m_pchBLReconFile, false, 8, g_uiBitDepth + g_uiBitIncrement ); // read mode
#endif
  // create & initialize internal classes
  xCreateDecLib();
  xInitDecLib  ();
  for(int i=0; i<MAX_LAYERS; i++)
    m_iPOCLastDisplay[i] += m_iSkipFrame;      // set the last displayed POC correctly for skip forward.

  // main decoder loop
  bool recon_opened[MAX_LAYERS] = {false}; // reconstruction file not yet opened. (must be performed after SPS is seen)
  Int iPrevLayer=0;
#if AVC_BL_ENABLE
  TComPicYuv*  pcPicYuvBLRec = NULL;
#endif
  while (!!bitstreamFile)
  {
    /* location serves to work around a design fault in the decoder, whereby
     * the process of reading a new slice that is the first slice of a new frame
     * requires the TDecTop::decode() method to be called again with the same
     * nal unit. */
    streampos location = bitstreamFile.tellg();
    AnnexBStats stats = AnnexBStats();
    bool bPreviousPictureDecoded = false;

    vector<uint8_t> nalUnit;
    InputNALUnit nalu;
    byteStreamNALUnit(bytestream, nalUnit, stats);
    Int iCurrLayer=0;

    // call actual decoding function
    bool bNewPicture = false;
    if (nalUnit.empty())
    {
      /* this can happen if the following occur:
       *  - empty input file
       *  - two back-to-back start_code_prefixes
       *  - start_code_prefix immediately followed by EOF
       */
      fprintf(stderr, "Warning: Attempt to decode an empty NAL unit\n");
    }
    else
    {
      read(nalu, nalUnit);
      iCurrLayer = nalu.m_svc_layer;
      if(m_iOutputInfo == 0 && iCurrLayer==1)
      {
        fprintf(stdout, "Skip Nalu, Layer:%d, NaluType:%d\n", iCurrLayer, nalu.m_nalUnitType);
        if(!bitstreamFile)
        {
          m_cTDecTop[0].executeDeblockAndAlf(uiPOC, pcListPic[0], m_iSkipFrame, m_iPOCLastDisplay[0]);
        }
        continue;
      }
/*#if TARGET_DECLAYERID_SET //SHM: need double check!
      if( (m_iMaxTemporalLayer >= 0 && nalu.m_temporalId > m_iMaxTemporalLayer) || !isNaluWithinTargetDecLayerIdSet(&nalu)  )
#else*/
      if(m_iMaxTemporalLayer >= 0 && nalu.m_temporalId > m_iMaxTemporalLayer)
//#endif
      {
        if(bPreviousPictureDecoded)
        {
          bNewPicture = true;
          bPreviousPictureDecoded = false;
        }
        else
        {
          bNewPicture = false;
        }
      }
      else
      {
        if(iCurrLayer != iPrevLayer)
        {
          int iLayer = iCurrLayer;
          iCurrLayer = iPrevLayer;
          iPrevLayer = iLayer;
          bNewPicture = true;
        }
        else
#if AVC_BL_ENABLE
        {
          // read base layer recon YUV
          if (nalu.isSlice())
          {
            m_cTDecTop[BASE_LAYER].xGetAvcBuffer(pcPicYuvBLRec);
            Int aiPad[2] = {0, 0};            
            if (pcPicYuvBLRec->getHeight()==544) aiPad[1] = 4;  // yong: SPS is 544, the base layer recon YUV is 960x544 (from HM), but avc recon is 960x540 (from JM)
            if (m_cTDecTop[iCurrLayer].getPrevPOC() != m_cTDecTop[iCurrLayer].getLastReadPOC())
              m_cTVideoIOYuvBLReconFile.read( pcPicYuvBLRec, aiPad, m_cTDecTop[iCurrLayer].getPrevPOC());
            m_cTDecTop[iCurrLayer].setLastReadPOC(m_cTDecTop[iCurrLayer].getPrevPOC());
            numNalRead ++;
          }
           bNewPicture = m_cTDecTop[iCurrLayer].decode(nalu, m_iSkipFrame, m_iPOCLastDisplay[iCurrLayer]);
        }
#else
           bNewPicture = m_cTDecTop[iCurrLayer].decode(nalu, m_iSkipFrame, m_iPOCLastDisplay[iCurrLayer]);
#endif

        if (bNewPicture)
        {
          bitstreamFile.clear();
          /* location points to the current nalunit payload[1] due to the
           * need for the annexB parser to read three extra bytes.
           * [1] except for the first NAL unit in the file
           *     (but bNewPicture doesn't happen then) */
          bitstreamFile.seekg(location-streamoff(3));
          bytestream.reset();
        }
        bPreviousPictureDecoded = true; 
      }
    }
    if (bNewPicture || !bitstreamFile)
    {
      m_cTDecTop[iCurrLayer].executeDeblockAndAlf(uiPOC, pcListPic[iCurrLayer], m_iSkipFrame, m_iPOCLastDisplay[iCurrLayer]);
    }

    if( pcListPic[iCurrLayer] )
    {
      if ( m_pchReconFile[iCurrLayer] && !recon_opened[iCurrLayer] )
      {
        if ( m_outputBitDepth == 0 )
        {
          m_outputBitDepth = g_uiBitDepth + g_uiBitIncrement;
        }

        m_cTVideoIOYuvReconFile[iCurrLayer].open( m_pchReconFile[iCurrLayer], true, m_outputBitDepth, g_uiBitDepth + g_uiBitIncrement ); // write mode
        recon_opened[iCurrLayer] = true;
      }
      if (bNewPicture && 
           (   nalu.m_nalUnitType == NAL_UNIT_CODED_SLICE_IDR
#if SUPPORT_FOR_RAP_N_LP
            || nalu.m_nalUnitType == NAL_UNIT_CODED_SLICE_IDR_N_LP
            || nalu.m_nalUnitType == NAL_UNIT_CODED_SLICE_BLA_N_LP
#endif
            || nalu.m_nalUnitType == NAL_UNIT_CODED_SLICE_BLANT
            || nalu.m_nalUnitType == NAL_UNIT_CODED_SLICE_BLA ) )
      {
        xFlushOutput( pcListPic[iCurrLayer], iCurrLayer );
      }
      // write reconstruction to file
      if(bNewPicture)
      {
        xWriteOutput( pcListPic[iCurrLayer], nalu.m_temporalId, iCurrLayer );
      }
    }
  }
  
  for(int i=0; i<MAX_LAYERS; i++)
  {
    if(pcListPic[i])
    {
     xFlushOutput( pcListPic[i], i );
     // delete buffers
     m_cTDecTop[i].deletePicBuffer();
    }
  }
#if AVC_BL_ENABLE
  pcPicYuvBLRec->destroy();
#endif
  // destroy internal classes
  xDestroyDecLib();
}
#else
Void TAppDecTop::decode()
{
  UInt                uiPOC;
  TComList<TComPic*>* pcListPic = NULL;

  ifstream bitstreamFile(m_pchBitstreamFile, ifstream::in | ifstream::binary);
  if (!bitstreamFile)
  {
    fprintf(stderr, "\nfailed to open bitstream file `%s' for reading\n", m_pchBitstreamFile);
    exit(EXIT_FAILURE);
  }

  InputByteStream bytestream(bitstreamFile);

  // create & initialize internal classes
  xCreateDecLib();
  xInitDecLib  ();
  m_iPOCLastDisplay += m_iSkipFrame;      // set the last displayed POC correctly for skip forward.

  // main decoder loop
  bool recon_opened = false; // reconstruction file not yet opened. (must be performed after SPS is seen)

  while (!!bitstreamFile)
  {
    /* location serves to work around a design fault in the decoder, whereby
     * the process of reading a new slice that is the first slice of a new frame
     * requires the TDecTop::decode() method to be called again with the same
     * nal unit. */
    streampos location = bitstreamFile.tellg();
    AnnexBStats stats = AnnexBStats();
    bool bPreviousPictureDecoded = false;

    vector<uint8_t> nalUnit;
    InputNALUnit nalu;
    byteStreamNALUnit(bytestream, nalUnit, stats);

    // call actual decoding function
    bool bNewPicture = false;
    if (nalUnit.empty())
    {
      /* this can happen if the following occur:
       *  - empty input file
       *  - two back-to-back start_code_prefixes
       *  - start_code_prefix immediately followed by EOF
       */
      fprintf(stderr, "Warning: Attempt to decode an empty NAL unit\n");
    }
    else
    {
      read(nalu, nalUnit);
#if TARGET_DECLAYERID_SET
      if( (m_iMaxTemporalLayer >= 0 && nalu.m_temporalId > m_iMaxTemporalLayer) || !isNaluWithinTargetDecLayerIdSet(&nalu)  )
#else
      if(m_iMaxTemporalLayer >= 0 && nalu.m_temporalId > m_iMaxTemporalLayer)
#endif
      {
        if(bPreviousPictureDecoded)
        {
          bNewPicture = true;
          bPreviousPictureDecoded = false;
        }
        else
        {
          bNewPicture = false;
        }
      }
      else
      {
        bNewPicture = m_cTDecTop.decode(nalu, m_iSkipFrame, m_iPOCLastDisplay);
        if (bNewPicture)
        {
          bitstreamFile.clear();
          /* location points to the current nalunit payload[1] due to the
           * need for the annexB parser to read three extra bytes.
           * [1] except for the first NAL unit in the file
           *     (but bNewPicture doesn't happen then) */
          bitstreamFile.seekg(location-streamoff(3));
          bytestream.reset();
        }
        bPreviousPictureDecoded = true; 
      }
    }
    if (bNewPicture || !bitstreamFile)
    {
      m_cTDecTop.executeDeblockAndAlf(uiPOC, pcListPic, m_iSkipFrame, m_iPOCLastDisplay);
    }

    if( pcListPic )
    {
      if ( m_pchReconFile && !recon_opened )
      {
        if ( m_outputBitDepth == 0 )
        {
          m_outputBitDepth = g_uiBitDepth + g_uiBitIncrement;
        }

        m_cTVideoIOYuvReconFile.open( m_pchReconFile, true, m_outputBitDepth, g_uiBitDepth + g_uiBitIncrement ); // write mode
        recon_opened = true;
      }
      if ( bNewPicture && 
           (   nalu.m_nalUnitType == NAL_UNIT_CODED_SLICE_IDR
#if SUPPORT_FOR_RAP_N_LP
            || nalu.m_nalUnitType == NAL_UNIT_CODED_SLICE_IDR_N_LP
            || nalu.m_nalUnitType == NAL_UNIT_CODED_SLICE_BLA_N_LP
#endif
            || nalu.m_nalUnitType == NAL_UNIT_CODED_SLICE_BLANT
            || nalu.m_nalUnitType == NAL_UNIT_CODED_SLICE_BLA ) )
      {
        xFlushOutput( pcListPic );
      }
      // write reconstruction to file
      if(bNewPicture)
      {
        xWriteOutput( pcListPic, nalu.m_temporalId );
      }
    }
  }
  
  xFlushOutput( pcListPic );
  // delete buffers
  m_cTDecTop.deletePicBuffer();
  
  // destroy internal classes
  xDestroyDecLib();
}
#endif
// ====================================================================================================================
// Protected member functions
// ====================================================================================================================
#ifdef HEVC_SVC
Void TAppDecTop::xCreateDecLib()
{
  // initialize ROM
  initROM();

  // create decoder class
  for(Int i=0; i<MAX_LAYERS; i++)
  {
   m_cTDecTop[i].create();
   m_cTDecTop[i].setLayerIdx(i);
#if AVC_BL_ENABLE
   m_cTDecTop[i].setPrevPOC(0);
   m_cTDecTop[i].setLastReadPOC(-1);
#endif
  }
  for(Int i=0; i<MAX_LAYERS; i++)
  {
    for(Int j=0; j<MAX_LAYERS; j++)
     m_cTDecTop[i].setLayerDecoder(j, &(m_cTDecTop[j]));
  }
}

Void TAppDecTop::xDestroyDecLib()
{
  //report;
  for(int i=0; i<(m_iOutputInfo==0? 1: MAX_LAYERS); i++)
  {
    m_cTDecTop[i].report();
  }

  //end;
  for(int i=0; i<MAX_LAYERS; i++)
  {
    if ( m_pchReconFile[i] )
    {
      m_cTVideoIOYuvReconFile[i]. close();
    }
#if AVC_BL_ENABLE
    if( m_pchBLReconFile )
      m_cTVideoIOYuvBLReconFile.close();
#endif
    // destroy decoder class
    m_cTDecTop[i].destroy();
  }
  // destroy ROM
  destroyROM();
}

Void TAppDecTop::xInitDecLib()
{
  // initialize decoder class
  for(int i=0; i<MAX_LAYERS; i++)
  {
   m_cTDecTop[i].init();
   m_cTDecTop[i].setDecodedPictureHashSEIEnabled(m_decodedPictureHashSEIEnabled);

  }
}

#else
Void TAppDecTop::xCreateDecLib()
{
  // create decoder class
  m_cTDecTop.create();
}

Void TAppDecTop::xDestroyDecLib()
{
  if ( m_pchReconFile )
  {
    m_cTVideoIOYuvReconFile. close();
  }
  
  // destroy decoder class
  m_cTDecTop.destroy();
}

Void TAppDecTop::xInitDecLib()
{
  // initialize decoder class
  m_cTDecTop.init();
  m_cTDecTop.setDecodedPictureHashSEIEnabled(m_decodedPictureHashSEIEnabled);
}
#endif

/** \param pcListPic list of pictures to be written to file
    \todo            DYN_REF_FREE should be revised
 */
#ifdef HEVC_SVC
Void TAppDecTop::xWriteOutput( TComList<TComPic*>* pcListPic, UInt tId, Int iLayerIdx )
#else
Void TAppDecTop::xWriteOutput( TComList<TComPic*>* pcListPic, UInt tId )
#endif
{
  TComList<TComPic*>::iterator iterPic   = pcListPic->begin();
  Int not_displayed = 0;

  while (iterPic != pcListPic->end())
  {
    TComPic* pcPic = *(iterPic);
#ifdef HEVC_SVC
    if(pcPic->getOutputMark() && pcPic->getPOC() > m_iPOCLastDisplay[iLayerIdx])
#else
    if(pcPic->getOutputMark() && pcPic->getPOC() > m_iPOCLastDisplay)
#endif
    {
       not_displayed++;
    }
    iterPic++;
  }
  iterPic   = pcListPic->begin();
  
  while (iterPic != pcListPic->end())
  {
    TComPic* pcPic = *(iterPic);
    TComSPS *sps = pcPic->getSlice(0)->getSPS();
#ifdef HEVC_SVC
    if ( pcPic->getOutputMark() && (not_displayed >  pcPic->getSlice(0)->getSPS()->getNumReorderPics(tId) && pcPic->getPOC() > m_iPOCLastDisplay[iLayerIdx]))
#else
    if ( pcPic->getOutputMark() && (not_displayed >  pcPic->getSlice(0)->getSPS()->getNumReorderPics(tId) && pcPic->getPOC() > m_iPOCLastDisplay))
#endif
    {
      // write to file
       not_displayed--;
#ifdef HEVC_SVC
      if ( m_pchReconFile[iLayerIdx] )
#else
      if ( m_pchReconFile )
#endif
      {
#ifdef HEVC_SVC
        m_cTVideoIOYuvReconFile[iLayerIdx].write( pcPic->getPicYuvRec(), sps->getPicCropLeftOffset(), sps->getPicCropRightOffset(), sps->getPicCropTopOffset(), sps->getPicCropBottomOffset() );
#else
        m_cTVideoIOYuvReconFile.write( pcPic->getPicYuvRec(), sps->getPicCropLeftOffset(), sps->getPicCropRightOffset(), sps->getPicCropTopOffset(), sps->getPicCropBottomOffset() );
#endif
      }
      
      // update POC of display order
#ifdef HEVC_SVC
      m_iPOCLastDisplay[iLayerIdx] = pcPic->getPOC();
#else
      m_iPOCLastDisplay = pcPic->getPOC();
#endif      
      // erase non-referenced picture in the reference picture list after display
      if ( !pcPic->getSlice(0)->isReferenced() && pcPic->getReconMark() == true )
      {
#if !DYN_REF_FREE
        pcPic->setReconMark(false);
        
        // mark it should be extended later
        pcPic->getPicYuvRec()->setBorderExtension( false );
        
#else
        pcPic->destroy();
        pcListPic->erase( iterPic );
        iterPic = pcListPic->begin(); // to the beginning, non-efficient way, have to be revised!
        continue;
#endif
      }
      pcPic->setOutputMark(false);
    }
    
    iterPic++;
  }
}

/** \param pcListPic list of pictures to be written to file
    \todo            DYN_REF_FREE should be revised
 */
#ifdef HEVC_SVC
Void TAppDecTop::xFlushOutput( TComList<TComPic*>* pcListPic, Int iLayerIdx )
#else
Void TAppDecTop::xFlushOutput( TComList<TComPic*>* pcListPic )
#endif
{
  if(!pcListPic)
  {
    return;
  } 
  TComList<TComPic*>::iterator iterPic   = pcListPic->begin();

  iterPic   = pcListPic->begin();
  
  while (iterPic != pcListPic->end())
  {
    TComPic* pcPic = *(iterPic);
    TComSPS *sps = pcPic->getSlice(0)->getSPS();

    if ( pcPic->getOutputMark() )
    {
      // write to file
#ifdef HEVC_SVC
      if ( m_pchReconFile[iLayerIdx] )
      {
        m_cTVideoIOYuvReconFile[iLayerIdx].write( pcPic->getPicYuvRec(), sps->getPicCropLeftOffset(), sps->getPicCropRightOffset(), sps->getPicCropTopOffset(), sps->getPicCropBottomOffset() );
      }
#else
      if ( m_pchReconFile )
      {
        m_cTVideoIOYuvReconFile.write( pcPic->getPicYuvRec(), sps->getPicCropLeftOffset(), sps->getPicCropRightOffset(), sps->getPicCropTopOffset(), sps->getPicCropBottomOffset() );
      }
#endif
      // update POC of display order
#ifdef HEVC_SVC
      m_iPOCLastDisplay[iLayerIdx] = pcPic->getPOC();
#else
      m_iPOCLastDisplay = pcPic->getPOC();
#endif      
      // erase non-referenced picture in the reference picture list after display
      if ( !pcPic->getSlice(0)->isReferenced() && pcPic->getReconMark() == true )
      {
#if !DYN_REF_FREE
        pcPic->setReconMark(false);
        
        // mark it should be extended later
        pcPic->getPicYuvRec()->setBorderExtension( false );
        
#else
        pcPic->destroy();
        pcListPic->erase( iterPic );
        iterPic = pcListPic->begin(); // to the beginning, non-efficient way, have to be revised!
        continue;
#endif
      }
      pcPic->setOutputMark(false);
    }
#if !DYN_REF_FREE
    if(pcPic)
    {
      pcPic->destroy();
      delete pcPic;
      pcPic = NULL;
    }
#endif    
    iterPic++;
  }
  pcListPic->clear();
#ifdef HEVC_SVC
  m_iPOCLastDisplay[iLayerIdx] = -MAX_INT;
#else
  m_iPOCLastDisplay = -MAX_INT;
#endif
}

#if TARGET_DECLAYERID_SET
/** \param nalu Input nalu to check whether its LayerId is within targetDecLayerIdSet
 */
Bool TAppDecTop::isNaluWithinTargetDecLayerIdSet( InputNALUnit* nalu )
{
  if ( m_targetDecLayerIdSet.size() == 0 ) // By default, the set is empty, meaning all LayerIds are allowed
  {
    return true;
  }
  for (std::vector<Int>::iterator it = m_targetDecLayerIdSet.begin(); it != m_targetDecLayerIdSet.end(); it++)
  {
    if ( nalu->m_reservedZero6Bits == (*it) )
    {
      return true;
    }
  }
  return false;
}
#endif

//! \}
