/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TAppEncTop.cpp
    \brief    Encoder application class
*/

#include <list>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <assert.h>

#include "TAppEncTop.h"
#include "TLibEncoder/AnnexBwrite.h"

#ifdef HEVC_SVC
#include <math.h>
#endif

using namespace std;

//! \ingroup TAppEncoder
//! \{

// ====================================================================================================================
// Constructor / destructor / initialization / destroy
// ====================================================================================================================

TAppEncTop::TAppEncTop()
{
#ifdef HEVC_SVC
  for (Int i = 0; i < MAX_LAYERS; i++)
  {
    m_iFrameRcvd[i] = 0;
    m_totalBytes[i] = 0;
    m_essentialBytes[i] = 0; 
    }
#else
  m_iFrameRcvd = 0;
  m_totalBytes = 0;
  m_essentialBytes = 0;
#endif

#ifdef HEVC_SVC
  m_cIlpPic[0] = NULL;
#endif
}

TAppEncTop::~TAppEncTop()
{
#ifdef HEVC_SVC
    if (m_cIlpPic[0] != NULL) 
    { 
      std::free(m_cIlpPic[0]); 
      m_cIlpPic[0] = NULL; 
    } 
#endif
}

#ifdef HEVC_SVC
Void TAppEncTop::xInitLibCfg(Int layer)
{
  TComVPS vps;
  
  vps.setMaxTLayers                       ( m_maxTempLayer );
  vps.setMaxLayers                        ( 1 );
  for(Int i = 0; i < MAX_TLAYER; i++)
  {
    vps.setNumReorderPics                 ( m_numReorderPics[i], i );
    vps.setMaxDecPicBuffering             ( m_maxDecPicBuffering[i], i );
  }
  m_cTEncTop[layer].setVPS(&vps);
  m_cTEncTop[layer].setFrameRate                    ( m_iFrameRate );
  m_cTEncTop[layer].setFrameSkip                    ( m_FrameSkip );
  m_cTEncTop[layer].setSourceWidth                  ( m_iSourceWidth );
  m_cTEncTop[layer].setSourceHeight                 ( m_iSourceHeight );
  m_cTEncTop[layer].setCroppingMode                 ( m_croppingMode );
  m_cTEncTop[layer].setCropLeft                     ( m_cropLeft );
  m_cTEncTop[layer].setCropRight                    ( m_cropRight );
  m_cTEncTop[layer].setCropTop                      ( m_cropTop );
  m_cTEncTop[layer].setCropBottom                   ( m_cropBottom );
  m_cTEncTop[layer].setFrameToBeEncoded             ( m_iFrameToBeEncoded );

#ifdef HEVC_SVC
  m_cTEncTop[layer].setUpSampleFilter               ( m_upSampleFilter );
  //multipass pic enc
  if(layer == 0)
    m_cTEncTop[layer].setNumMulPassEnc            ( 1 );
  else
    m_cTEncTop[layer].setNumMulPassEnc            ( m_iNumMulPassPic );
#endif

#ifdef HEVC_SVC
  m_cTEncTop[layer].setTemporalResolution            (m_uiTemporalResolution);
  m_cTEncTop[layer].setCUWidth                       ( m_uiMaxCUWidth );
  m_cTEncTop[layer].setCUHeight                      ( m_uiMaxCUHeight );
  m_cTEncTop[layer].setCUDepth                       ( m_uiMaxCUDepth );
  m_cTEncTop[layer].setBitstreamFile                 ( m_pchBitstreamFile );
  if (m_pchReconFile)
  {
    m_cTEncTop[layer].setReconFile                     ( m_pchReconFile );
  }
#endif
  
  //====== Coding Structure ========
  m_cTEncTop[layer].setIntraPeriod                  ( m_iIntraPeriod );
  m_cTEncTop[layer].setDecodingRefreshType          ( m_iDecodingRefreshType );
  m_cTEncTop[layer].setGOPSize                      ( m_iGOPSize );
  m_cTEncTop[layer].setGopList                      ( m_GOPList );
  m_cTEncTop[layer].setExtraRPSs                    ( m_extraRPSs );
  for(Int i = 0; i < MAX_TLAYER; i++)
  {
    m_cTEncTop[layer].setNumReorderPics             ( m_numReorderPics[i], i );
    m_cTEncTop[layer].setMaxDecPicBuffering         ( m_maxDecPicBuffering[i], i );
  }
  for( UInt uiLoop = 0; uiLoop < MAX_TLAYER; ++uiLoop )
  {
    m_cTEncTop[layer].setLambdaModifier( uiLoop, m_adLambdaModifier[ uiLoop ] );
  }
  m_cTEncTop[layer].setQP                           ( m_iQP );

#ifdef HEVC_SVC
  m_cTEncTop[layer].setQP_ELI                       ( m_iQP_ELI );
#endif
  
  m_cTEncTop[layer].setPad                          ( m_aiPad );

#ifdef HEVC_SVC
  if (layer == 0)
  {
      m_cTEncTop[layer].setIlpMode                    ( 0); 
  }
  else
  {
      m_cTEncTop[layer].setIlpMode                    ( m_ilpModeConfig); 
  }
  m_cTEncTop[layer].setIlpReferenceLayer( m_ilpReferenceLayerConfig); 
  m_cTEncTop[layer].setIdenticalTemporalRps(m_IdenticalTemporalRpsConfig);
#endif
    
  m_cTEncTop[layer].setMaxTempLayer                 ( m_maxTempLayer );
#if !REMOVE_NSQT
  m_cTEncTop[layer].setUseNSQT( m_enableNSQT );
#endif
  m_cTEncTop[layer].setUseAMP( m_enableAMP );
  
  //===== Slice ========
  
  //====== Loop/Deblock Filter ========
  m_cTEncTop[layer].setLoopFilterDisable            ( m_bLoopFilterDisable       );
  m_cTEncTop[layer].setLoopFilterOffsetInPPS        ( m_loopFilterOffsetInPPS );
  m_cTEncTop[layer].setLoopFilterBetaOffset         ( m_loopFilterBetaOffsetDiv2  );
  m_cTEncTop[layer].setLoopFilterTcOffset           ( m_loopFilterTcOffsetDiv2    );
  m_cTEncTop[layer].setDeblockingFilterControlPresent( m_DeblockingFilterControlPresent);

  //====== Motion search ========
  m_cTEncTop[layer].setFastSearch                   ( m_iFastSearch  );
  m_cTEncTop[layer].setSearchRange                  ( m_iSearchRange );
  m_cTEncTop[layer].setBipredSearchRange            ( m_bipredSearchRange );

  //====== Quality control ========
  m_cTEncTop[layer].setMaxDeltaQP                   ( m_iMaxDeltaQP  );
  m_cTEncTop[layer].setMaxCuDQPDepth                ( m_iMaxCuDQPDepth  );

  m_cTEncTop[layer].setChromaCbQpOffset               ( m_cbQpOffset     );
  m_cTEncTop[layer].setChromaCrQpOffset            ( m_crQpOffset  );

#if ADAPTIVE_QP_SELECTION
  m_cTEncTop[layer].setUseAdaptQpSelect             ( m_bUseAdaptQpSelect   );
#endif

  Int lowestQP;
  lowestQP =  - ( (Int)(6*(g_uiBitDepth + g_uiBitIncrement - 8)) );

  if ((m_iMaxDeltaQP == 0 ) && (m_iQP == lowestQP) && (m_useLossless == true))
  {
    m_bUseAdaptiveQP = false;
  }
  m_cTEncTop[layer].setUseAdaptiveQP                ( m_bUseAdaptiveQP  );
  m_cTEncTop[layer].setQPAdaptationRange            ( m_iQPAdaptationRange );
  
  //====== Tool list ========
  m_cTEncTop[layer].setUseSBACRD                    ( m_bUseSBACRD   );
  m_cTEncTop[layer].setDeltaQpRD                    ( m_uiDeltaQpRD  );
  m_cTEncTop[layer].setUseASR                       ( m_bUseASR      );
  m_cTEncTop[layer].setUseHADME                     ( m_bUseHADME    );
#if !REMOVE_ALF
  m_cTEncTop[layer].setUseALF                       ( m_bUseALF      );
#endif
  m_cTEncTop[layer].setUseLossless                  ( m_useLossless );
  m_cTEncTop[layer].setUseLComb                     ( m_bUseLComb    );
  m_cTEncTop[layer].setdQPs                         ( m_aidQP        );
  m_cTEncTop[layer].setUseRDOQ                      ( m_bUseRDOQ     );
  m_cTEncTop[layer].setQuadtreeTULog2MaxSize        ( m_uiQuadtreeTULog2MaxSize );
  m_cTEncTop[layer].setQuadtreeTULog2MinSize        ( m_uiQuadtreeTULog2MinSize );
  m_cTEncTop[layer].setQuadtreeTUMaxDepthInter      ( m_uiQuadtreeTUMaxDepthInter );
  m_cTEncTop[layer].setQuadtreeTUMaxDepthIntra      ( m_uiQuadtreeTUMaxDepthIntra );
  m_cTEncTop[layer].setUseFastEnc                   ( m_bUseFastEnc  );
  m_cTEncTop[layer].setUseEarlyCU                   ( m_bUseEarlyCU  ); 
  m_cTEncTop[layer].setUseFastDecisionForMerge      ( m_useFastDecisionForMerge  );
  m_cTEncTop[layer].setUseCbfFastMode            ( m_bUseCbfFastMode  );
  m_cTEncTop[layer].setUseEarlySkipDetection            ( m_useEarlySkipDetection );

#if !REMOVE_LMCHROMA
  m_cTEncTop.setUseLMChroma                  ( m_bUseLMChroma );
#endif
  m_cTEncTop[layer].setUseTransformSkip             ( m_useTransformSkip      );
  m_cTEncTop[layer].setUseTransformSkipFast         ( m_useTransformSkipFast  );
  m_cTEncTop[layer].setUseConstrainedIntraPred      ( m_bUseConstrainedIntraPred );
  m_cTEncTop[layer].setPCMLog2MinSize          ( m_uiPCMLog2MinSize);
  m_cTEncTop[layer].setUsePCM                       ( m_usePCM );
  m_cTEncTop[layer].setPCMLog2MaxSize               ( m_pcmLog2MaxSize);
  m_cTEncTop[layer].setMaxNumMergeCand              ( m_maxNumMergeCand );
  

  //====== Weighted Prediction ========
  m_cTEncTop[layer].setUseWP                   ( m_bUseWeightPred      );
  m_cTEncTop[layer].setWPBiPred                ( m_useWeightedBiPred   );
  //====== Parallel Merge Estimation ========
  m_cTEncTop[layer].setLog2ParallelMergeLevelMinus2 ( m_log2ParallelMergeLevel - 2 );

  //====== Slice ========
  m_cTEncTop[layer].setSliceMode               ( m_iSliceMode                );
  m_cTEncTop[layer].setSliceArgument           ( m_iSliceArgument            );

  //====== Dependent Slice ========
  m_cTEncTop[layer].setDependentSliceMode        ( m_iDependentSliceMode         );
  m_cTEncTop[layer].setDependentSliceArgument    ( m_iDependentSliceArgument     );
#if DEPENDENT_SLICES
#if TILES_WPP_ENTROPYSLICES_FLAGS
  m_cTEncTop[layer].setEntropySliceEnabledFlag   ( m_entropySliceEnabledFlag );
#else
  m_cTEncTop.setCabacIndependentFlag      ( m_bCabacIndependentFlag   );
#endif
#endif
  int iNumPartInCU = 1<<(m_uiMaxCUDepth<<1);
  if(m_iDependentSliceMode==SHARP_FIXED_NUMBER_OF_LCU_IN_DEPENDENT_SLICE)
  {
#if REMOVE_FGS
    m_cTEncTop[layer].setDependentSliceArgument ( m_iDependentSliceArgument * iNumPartInCU );
#else
    m_cTEncTop.setDependentSliceArgument ( m_iDependentSliceArgument * ( iNumPartInCU >> ( m_iSliceGranularity << 1 ) ) );
#endif
  }
  if(m_iSliceMode==AD_HOC_SLICES_FIXED_NUMBER_OF_LCU_IN_SLICE)
  {
#if REMOVE_FGS
    m_cTEncTop[layer].setSliceArgument ( m_iSliceArgument * iNumPartInCU );
#else
    m_cTEncTop.setSliceArgument ( m_iSliceArgument * ( iNumPartInCU >> ( m_iSliceGranularity << 1 ) ) );
#endif
  }
  if(m_iSliceMode==AD_HOC_SLICES_FIXED_NUMBER_OF_TILES_IN_SLICE)
  {
    m_cTEncTop[layer].setSliceArgument ( m_iSliceArgument );
  }
  
#if !REMOVE_FGS
  m_cTEncTop.setSliceGranularity        ( m_iSliceGranularity         );
#endif
  if(m_iSliceMode == 0 )
  {
    m_bLFCrossSliceBoundaryFlag = true;
  }
  m_cTEncTop[layer].setLFCrossSliceBoundaryFlag( m_bLFCrossSliceBoundaryFlag );
  m_cTEncTop[layer].setUseSAO ( m_bUseSAO );
  m_cTEncTop[layer].setMaxNumOffsetsPerPic (m_maxNumOffsetsPerPic);

#if SAO_LCU_BOUNDARY
  m_cTEncTop[layer].setSaoLcuBoundary (m_saoLcuBoundary);
#endif
  m_cTEncTop[layer].setSaoLcuBasedOptimization (m_saoLcuBasedOptimization);
  m_cTEncTop[layer].setPCMInputBitDepthFlag  ( m_bPCMInputBitDepthFlag); 
  m_cTEncTop[layer].setPCMFilterDisableFlag  ( m_bPCMFilterDisableFlag); 

  m_cTEncTop[layer].setDecodedPictureHashSEIEnabled(m_decodePictureHashSEIEnabled);
#if RECOVERY_POINT_SEI
  m_cTEncTop[layer].setRecoveryPointSEIEnabled( m_recoveryPointSEIEnabled );
#endif
#if BUFFERING_PERIOD_AND_TIMING_SEI
  m_cTEncTop[layer].setBufferingPeriodSEIEnabled( m_bufferingPeriodSEIEnabled );
  m_cTEncTop[layer].setPictureTimingSEIEnabled( m_pictureTimingSEIEnabled );
#endif
  m_cTEncTop[layer].setUniformSpacingIdr          ( m_iUniformSpacingIdr );
  m_cTEncTop[layer].setNumColumnsMinus1           ( m_iNumColumnsMinus1 );
  m_cTEncTop[layer].setNumRowsMinus1              ( m_iNumRowsMinus1 );
  if(m_iUniformSpacingIdr==0)
  {
    m_cTEncTop[layer].setColumnWidth              ( m_pchColumnWidth );
    m_cTEncTop[layer].setRowHeight                ( m_pchRowHeight );
  }
  m_cTEncTop[layer].xCheckGSParameters();
  Int uiTilesCount          = (m_iNumRowsMinus1+1) * (m_iNumColumnsMinus1+1);
  if(uiTilesCount == 1)
  {
    m_bLFCrossTileBoundaryFlag = true; 
  }
  m_cTEncTop[layer].setLFCrossTileBoundaryFlag( m_bLFCrossTileBoundaryFlag );
  m_cTEncTop[layer].setWaveFrontSynchro           ( m_iWaveFrontSynchro );
  m_cTEncTop[layer].setWaveFrontSubstreams        ( m_iWaveFrontSubstreams );
  m_cTEncTop[layer].setTMVPModeId ( m_TMVPModeId );
  m_cTEncTop[layer].setUseScalingListId           ( m_useScalingListId  );
  m_cTEncTop[layer].setScalingListFile            ( m_scalingListFile   );
  m_cTEncTop[layer].setSignHideFlag(m_signHideFlag);
#if !REMOVE_ALF
  m_cTEncTop.setALFLowLatencyEncoding( m_alfLowLatencyEncoding );
#endif
  m_cTEncTop[layer].setUseRateCtrl     ( m_enableRateCtrl);
  m_cTEncTop[layer].setTargetBitrate   ( m_targetBitrate);
  m_cTEncTop[layer].setNumLCUInUnit    ( m_numLCUInUnit);
  m_cTEncTop[layer].setTransquantBypassEnableFlag(m_TransquantBypassEnableFlag);
  m_cTEncTop[layer].setCUTransquantBypassFlagValue(m_CUTransquantBypassFlagValue);
#if RECALCULATE_QP_ACCORDING_LAMBDA
  m_cTEncTop[layer].setUseRecalculateQPAccordingToLambda( m_recalculateQPAccordingToLambda );
#endif
#if ACTIVE_PARAMETER_SETS_SEI_MESSAGE
  m_cTEncTop[layer].setActiveParameterSetsSEIEnabled ( m_activeParameterSetsSEIEnabled ); 
#endif 
#if SUPPORT_FOR_VUI
  m_cTEncTop[layer].setVuiParametersPresentFlag( m_vuiParametersPresentFlag );
  m_cTEncTop[layer].setAspectRatioIdc( m_aspectRatioIdc );
  m_cTEncTop[layer].setSarWidth( m_sarWidth );
  m_cTEncTop[layer].setSarHeight( m_sarHeight );
  m_cTEncTop[layer].setOverscanInfoPresentFlag( m_overscanInfoPresentFlag );
  m_cTEncTop[layer].setOverscanAppropriateFlag( m_overscanAppropriateFlag );
  m_cTEncTop[layer].setVideoSignalTypePresentFlag( m_videoSignalTypePresentFlag );
  m_cTEncTop[layer].setVideoFormat( m_videoFormat );
  m_cTEncTop[layer].setVideoFullRangeFlag( m_videoFullRangeFlag );
  m_cTEncTop[layer].setColourDescriptionPresentFlag( m_colourDescriptionPresentFlag );
  m_cTEncTop[layer].setColourPrimaries( m_colourPrimaries );
  m_cTEncTop[layer].setTransferCharacteristics( m_transferCharacteristics );
  m_cTEncTop[layer].setMatrixCoefficients( m_matrixCoefficients );
  m_cTEncTop[layer].setChromaLocInfoPresentFlag( m_chromaLocInfoPresentFlag );
  m_cTEncTop[layer].setChromaSampleLocTypeTopField( m_chromaSampleLocTypeTopField );
  m_cTEncTop[layer].setChromaSampleLocTypeBottomField( m_chromaSampleLocTypeBottomField );
  m_cTEncTop[layer].setNeutralChromaIndicationFlag( m_neutralChromaIndicationFlag );
  m_cTEncTop[layer].setBitstreamRestrictionFlag( m_bitstreamRestrictionFlag );
  m_cTEncTop[layer].setTilesFixedStructureFlag( m_tilesFixedStructureFlag );
  m_cTEncTop[layer].setMotionVectorsOverPicBoundariesFlag( m_motionVectorsOverPicBoundariesFlag );
  m_cTEncTop[layer].setMaxBytesPerPicDenom( m_maxBytesPerPicDenom );
  m_cTEncTop[layer].setMaxBitsPerMinCuDenom( m_maxBitsPerMinCuDenom );
  m_cTEncTop[layer].setLog2MaxMvLengthHorizontal( m_log2MaxMvLengthHorizontal );
  m_cTEncTop[layer].setLog2MaxMvLengthVertical( m_log2MaxMvLengthVertical );
#endif
}

Void TAppEncTop::xCreateLib(Int layer)
{
  // Video I/O
  m_cTVideoIOYuvInputFile[layer].open( m_pchInputFile,     false, m_uiInputBitDepth, m_uiInternalBitDepth );  // read  mode
#if !HEVC_SVC_MEMLEAK_FIX
  free(m_pchInputFile);
#else
  sFree(m_pchInputFile);
#endif
  m_cTVideoIOYuvInputFile[layer].skipFrames(m_FrameSkip, m_iSourceWidth, m_iSourceHeight);

  if (m_pchReconFile)
    m_cTVideoIOYuvReconFile[layer].open(m_pchReconFile, true, m_uiOutputBitDepth, m_uiInternalBitDepth);  // write mode
  
  // Neo Decoder
  m_cTEncTop[layer].create();
}

Void TAppEncTop::xDestroyLib(Int layer)
{
  // Video I/O
  m_cTVideoIOYuvInputFile[layer].close();
  m_cTVideoIOYuvReconFile[layer].close();
  
  // Neo Decoder
  m_cTEncTop[layer].destroy();
}

Void TAppEncTop::xInitLib(Int layer)
{
  m_cTEncTop[layer].init(layer);
}
#else
Void TAppEncTop::xInitLibCfg()
{
  TComVPS vps;
  
  vps.setMaxTLayers                       ( m_maxTempLayer );
  vps.setMaxLayers                        ( 1 );
  for(Int i = 0; i < MAX_TLAYER; i++)
  {
    vps.setNumReorderPics                 ( m_numReorderPics[i], i );
    vps.setMaxDecPicBuffering             ( m_maxDecPicBuffering[i], i );
  }
  m_cTEncTop.setVPS(&vps);
  m_cTEncTop.setFrameRate                    ( m_iFrameRate );
  m_cTEncTop.setFrameSkip                    ( m_FrameSkip );
  m_cTEncTop.setSourceWidth                  ( m_iSourceWidth );
  m_cTEncTop.setSourceHeight                 ( m_iSourceHeight );
  m_cTEncTop.setCroppingMode                 ( m_croppingMode );
  m_cTEncTop.setCropLeft                     ( m_cropLeft );
  m_cTEncTop.setCropRight                    ( m_cropRight );
  m_cTEncTop.setCropTop                      ( m_cropTop );
  m_cTEncTop.setCropBottom                   ( m_cropBottom );
  m_cTEncTop.setFrameToBeEncoded             ( m_iFrameToBeEncoded );
  
  //====== Coding Structure ========
  m_cTEncTop.setIntraPeriod                  ( m_iIntraPeriod );
  m_cTEncTop.setDecodingRefreshType          ( m_iDecodingRefreshType );
  m_cTEncTop.setGOPSize                      ( m_iGOPSize );
  m_cTEncTop.setGopList                      ( m_GOPList );
  m_cTEncTop.setExtraRPSs                    ( m_extraRPSs );
  for(Int i = 0; i < MAX_TLAYER; i++)
  {
    m_cTEncTop.setNumReorderPics             ( m_numReorderPics[i], i );
    m_cTEncTop.setMaxDecPicBuffering         ( m_maxDecPicBuffering[i], i );
  }
  for( UInt uiLoop = 0; uiLoop < MAX_TLAYER; ++uiLoop )
  {
    m_cTEncTop.setLambdaModifier( uiLoop, m_adLambdaModifier[ uiLoop ] );
  }
  m_cTEncTop.setQP                           ( m_iQP );
  
  m_cTEncTop.setPad                          ( m_aiPad );
    
  m_cTEncTop.setMaxTempLayer                 ( m_maxTempLayer );
#if !REMOVE_NSQT
  m_cTEncTop.setUseNSQT( m_enableNSQT );
#endif
  m_cTEncTop.setUseAMP( m_enableAMP );
  
  //===== Slice ========
  
  //====== Loop/Deblock Filter ========
  m_cTEncTop.setLoopFilterDisable            ( m_bLoopFilterDisable       );
  m_cTEncTop.setLoopFilterOffsetInPPS        ( m_loopFilterOffsetInPPS );
  m_cTEncTop.setLoopFilterBetaOffset         ( m_loopFilterBetaOffsetDiv2  );
  m_cTEncTop.setLoopFilterTcOffset           ( m_loopFilterTcOffsetDiv2    );
  m_cTEncTop.setDeblockingFilterControlPresent( m_DeblockingFilterControlPresent);

  //====== Motion search ========
  m_cTEncTop.setFastSearch                   ( m_iFastSearch  );
  m_cTEncTop.setSearchRange                  ( m_iSearchRange );
  m_cTEncTop.setBipredSearchRange            ( m_bipredSearchRange );

  //====== Quality control ========
  m_cTEncTop.setMaxDeltaQP                   ( m_iMaxDeltaQP  );
  m_cTEncTop.setMaxCuDQPDepth                ( m_iMaxCuDQPDepth  );

  m_cTEncTop.setChromaCbQpOffset               ( m_cbQpOffset     );
  m_cTEncTop.setChromaCrQpOffset            ( m_crQpOffset  );

#if ADAPTIVE_QP_SELECTION
  m_cTEncTop.setUseAdaptQpSelect             ( m_bUseAdaptQpSelect   );
#endif

  Int lowestQP;
  lowestQP =  - ( (Int)(6*(g_uiBitDepth + g_uiBitIncrement - 8)) );

  if ((m_iMaxDeltaQP == 0 ) && (m_iQP == lowestQP) && (m_useLossless == true))
  {
    m_bUseAdaptiveQP = false;
  }
  m_cTEncTop.setUseAdaptiveQP                ( m_bUseAdaptiveQP  );
  m_cTEncTop.setQPAdaptationRange            ( m_iQPAdaptationRange );
  
  //====== Tool list ========
  m_cTEncTop.setUseSBACRD                    ( m_bUseSBACRD   );
  m_cTEncTop.setDeltaQpRD                    ( m_uiDeltaQpRD  );
  m_cTEncTop.setUseASR                       ( m_bUseASR      );
  m_cTEncTop.setUseHADME                     ( m_bUseHADME    );
#if !REMOVE_ALF
  m_cTEncTop.setUseALF                       ( m_bUseALF      );
#endif
  m_cTEncTop.setUseLossless                  ( m_useLossless );
  m_cTEncTop.setUseLComb                     ( m_bUseLComb    );
  m_cTEncTop.setdQPs                         ( m_aidQP        );
  m_cTEncTop.setUseRDOQ                      ( m_bUseRDOQ     );
  m_cTEncTop.setQuadtreeTULog2MaxSize        ( m_uiQuadtreeTULog2MaxSize );
  m_cTEncTop.setQuadtreeTULog2MinSize        ( m_uiQuadtreeTULog2MinSize );
  m_cTEncTop.setQuadtreeTUMaxDepthInter      ( m_uiQuadtreeTUMaxDepthInter );
  m_cTEncTop.setQuadtreeTUMaxDepthIntra      ( m_uiQuadtreeTUMaxDepthIntra );
  m_cTEncTop.setUseFastEnc                   ( m_bUseFastEnc  );
  m_cTEncTop.setUseEarlyCU                   ( m_bUseEarlyCU  ); 
  m_cTEncTop.setUseFastDecisionForMerge      ( m_useFastDecisionForMerge  );
  m_cTEncTop.setUseCbfFastMode            ( m_bUseCbfFastMode  );
  m_cTEncTop.setUseEarlySkipDetection            ( m_useEarlySkipDetection );

#if !REMOVE_LMCHROMA
  m_cTEncTop.setUseLMChroma                  ( m_bUseLMChroma );
#endif
  m_cTEncTop.setUseTransformSkip             ( m_useTransformSkip      );
  m_cTEncTop.setUseTransformSkipFast         ( m_useTransformSkipFast  );
  m_cTEncTop.setUseConstrainedIntraPred      ( m_bUseConstrainedIntraPred );
  m_cTEncTop.setPCMLog2MinSize          ( m_uiPCMLog2MinSize);
  m_cTEncTop.setUsePCM                       ( m_usePCM );
  m_cTEncTop.setPCMLog2MaxSize               ( m_pcmLog2MaxSize);
  m_cTEncTop.setMaxNumMergeCand              ( m_maxNumMergeCand );
  

  //====== Weighted Prediction ========
  m_cTEncTop.setUseWP                   ( m_bUseWeightPred      );
  m_cTEncTop.setWPBiPred                ( m_useWeightedBiPred   );
  //====== Parallel Merge Estimation ========
  m_cTEncTop.setLog2ParallelMergeLevelMinus2 ( m_log2ParallelMergeLevel - 2 );

  //====== Slice ========
  m_cTEncTop.setSliceMode               ( m_iSliceMode                );
  m_cTEncTop.setSliceArgument           ( m_iSliceArgument            );

  //====== Dependent Slice ========
  m_cTEncTop.setDependentSliceMode        ( m_iDependentSliceMode         );
  m_cTEncTop.setDependentSliceArgument    ( m_iDependentSliceArgument     );
#if DEPENDENT_SLICES
#if TILES_WPP_ENTROPYSLICES_FLAGS
  m_cTEncTop.setEntropySliceEnabledFlag   ( m_entropySliceEnabledFlag );
#else
  m_cTEncTop.setCabacIndependentFlag      ( m_bCabacIndependentFlag   );
#endif
#endif
  int iNumPartInCU = 1<<(m_uiMaxCUDepth<<1);
  if(m_iDependentSliceMode==SHARP_FIXED_NUMBER_OF_LCU_IN_DEPENDENT_SLICE)
  {
#if REMOVE_FGS
    m_cTEncTop.setDependentSliceArgument ( m_iDependentSliceArgument * iNumPartInCU );
#else
    m_cTEncTop.setDependentSliceArgument ( m_iDependentSliceArgument * ( iNumPartInCU >> ( m_iSliceGranularity << 1 ) ) );
#endif
  }
  if(m_iSliceMode==AD_HOC_SLICES_FIXED_NUMBER_OF_LCU_IN_SLICE)
  {
#if REMOVE_FGS
    m_cTEncTop.setSliceArgument ( m_iSliceArgument * iNumPartInCU );
#else
    m_cTEncTop.setSliceArgument ( m_iSliceArgument * ( iNumPartInCU >> ( m_iSliceGranularity << 1 ) ) );
#endif
  }
  if(m_iSliceMode==AD_HOC_SLICES_FIXED_NUMBER_OF_TILES_IN_SLICE)
  {
    m_cTEncTop.setSliceArgument ( m_iSliceArgument );
  }
  
#if !REMOVE_FGS
  m_cTEncTop.setSliceGranularity        ( m_iSliceGranularity         );
#endif
  if(m_iSliceMode == 0 )
  {
    m_bLFCrossSliceBoundaryFlag = true;
  }
  m_cTEncTop.setLFCrossSliceBoundaryFlag( m_bLFCrossSliceBoundaryFlag );
  m_cTEncTop.setUseSAO ( m_bUseSAO );
  m_cTEncTop.setMaxNumOffsetsPerPic (m_maxNumOffsetsPerPic);

#if SAO_LCU_BOUNDARY
  m_cTEncTop.setSaoLcuBoundary (m_saoLcuBoundary);
#endif
  m_cTEncTop.setSaoLcuBasedOptimization (m_saoLcuBasedOptimization);
  m_cTEncTop.setPCMInputBitDepthFlag  ( m_bPCMInputBitDepthFlag); 
  m_cTEncTop.setPCMFilterDisableFlag  ( m_bPCMFilterDisableFlag); 

  m_cTEncTop.setDecodedPictureHashSEIEnabled(m_decodePictureHashSEIEnabled);
#if RECOVERY_POINT_SEI
  m_cTEncTop.setRecoveryPointSEIEnabled( m_recoveryPointSEIEnabled );
#endif
#if BUFFERING_PERIOD_AND_TIMING_SEI
  m_cTEncTop.setBufferingPeriodSEIEnabled( m_bufferingPeriodSEIEnabled );
  m_cTEncTop.setPictureTimingSEIEnabled( m_pictureTimingSEIEnabled );
#endif
  m_cTEncTop.setUniformSpacingIdr          ( m_iUniformSpacingIdr );
  m_cTEncTop.setNumColumnsMinus1           ( m_iNumColumnsMinus1 );
  m_cTEncTop.setNumRowsMinus1              ( m_iNumRowsMinus1 );
  if(m_iUniformSpacingIdr==0)
  {
    m_cTEncTop.setColumnWidth              ( m_pchColumnWidth );
    m_cTEncTop.setRowHeight                ( m_pchRowHeight );
  }
  m_cTEncTop.xCheckGSParameters();
  Int uiTilesCount          = (m_iNumRowsMinus1+1) * (m_iNumColumnsMinus1+1);
  if(uiTilesCount == 1)
  {
    m_bLFCrossTileBoundaryFlag = true; 
  }
  m_cTEncTop.setLFCrossTileBoundaryFlag( m_bLFCrossTileBoundaryFlag );
  m_cTEncTop.setWaveFrontSynchro           ( m_iWaveFrontSynchro );
  m_cTEncTop.setWaveFrontSubstreams        ( m_iWaveFrontSubstreams );
  m_cTEncTop.setTMVPModeId ( m_TMVPModeId );
  m_cTEncTop.setUseScalingListId           ( m_useScalingListId  );
  m_cTEncTop.setScalingListFile            ( m_scalingListFile   );
  m_cTEncTop.setSignHideFlag(m_signHideFlag);
#if !REMOVE_ALF
  m_cTEncTop.setALFLowLatencyEncoding( m_alfLowLatencyEncoding );
#endif
  m_cTEncTop.setUseRateCtrl     ( m_enableRateCtrl);
  m_cTEncTop.setTargetBitrate   ( m_targetBitrate);
  m_cTEncTop.setNumLCUInUnit    ( m_numLCUInUnit);
  m_cTEncTop.setTransquantBypassEnableFlag(m_TransquantBypassEnableFlag);
  m_cTEncTop.setCUTransquantBypassFlagValue(m_CUTransquantBypassFlagValue);
#if RECALCULATE_QP_ACCORDING_LAMBDA
  m_cTEncTop.setUseRecalculateQPAccordingToLambda( m_recalculateQPAccordingToLambda );
#endif
#if ACTIVE_PARAMETER_SETS_SEI_MESSAGE
  m_cTEncTop.setActiveParameterSetsSEIEnabled ( m_activeParameterSetsSEIEnabled ); 
#endif 
#if SUPPORT_FOR_VUI
  m_cTEncTop.setVuiParametersPresentFlag( m_vuiParametersPresentFlag );
  m_cTEncTop.setAspectRatioIdc( m_aspectRatioIdc );
  m_cTEncTop.setSarWidth( m_sarWidth );
  m_cTEncTop.setSarHeight( m_sarHeight );
  m_cTEncTop.setOverscanInfoPresentFlag( m_overscanInfoPresentFlag );
  m_cTEncTop.setOverscanAppropriateFlag( m_overscanAppropriateFlag );
  m_cTEncTop.setVideoSignalTypePresentFlag( m_videoSignalTypePresentFlag );
  m_cTEncTop.setVideoFormat( m_videoFormat );
  m_cTEncTop.setVideoFullRangeFlag( m_videoFullRangeFlag );
  m_cTEncTop.setColourDescriptionPresentFlag( m_colourDescriptionPresentFlag );
  m_cTEncTop.setColourPrimaries( m_colourPrimaries );
  m_cTEncTop.setTransferCharacteristics( m_transferCharacteristics );
  m_cTEncTop.setMatrixCoefficients( m_matrixCoefficients );
  m_cTEncTop.setChromaLocInfoPresentFlag( m_chromaLocInfoPresentFlag );
  m_cTEncTop.setChromaSampleLocTypeTopField( m_chromaSampleLocTypeTopField );
  m_cTEncTop.setChromaSampleLocTypeBottomField( m_chromaSampleLocTypeBottomField );
  m_cTEncTop.setNeutralChromaIndicationFlag( m_neutralChromaIndicationFlag );
  m_cTEncTop.setBitstreamRestrictionFlag( m_bitstreamRestrictionFlag );
  m_cTEncTop.setTilesFixedStructureFlag( m_tilesFixedStructureFlag );
  m_cTEncTop.setMotionVectorsOverPicBoundariesFlag( m_motionVectorsOverPicBoundariesFlag );
  m_cTEncTop.setMaxBytesPerPicDenom( m_maxBytesPerPicDenom );
  m_cTEncTop.setMaxBitsPerMinCuDenom( m_maxBitsPerMinCuDenom );
  m_cTEncTop.setLog2MaxMvLengthHorizontal( m_log2MaxMvLengthHorizontal );
  m_cTEncTop.setLog2MaxMvLengthVertical( m_log2MaxMvLengthVertical );
#endif
}

Void TAppEncTop::xCreateLib()
{
  // Video I/O
  m_cTVideoIOYuvInputFile.open( m_pchInputFile,     false, m_uiInputBitDepth, m_uiInternalBitDepth );  // read  mode
  m_cTVideoIOYuvInputFile.skipFrames(m_FrameSkip, m_iSourceWidth - m_aiPad[0], m_iSourceHeight - m_aiPad[1]);

  if (m_pchReconFile)
    m_cTVideoIOYuvReconFile.open(m_pchReconFile, true, m_uiOutputBitDepth, m_uiInternalBitDepth);  // write mode
  
  // Neo Decoder
  m_cTEncTop.create();
}

Void TAppEncTop::xDestroyLib()
{
  // Video I/O
  m_cTVideoIOYuvInputFile.close();
  m_cTVideoIOYuvReconFile.close();
  
  // Neo Decoder
  m_cTEncTop.destroy();
}

Void TAppEncTop::xInitLib()
{
  m_cTEncTop.init();
}
#endif

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

/**
 - create internal class
 - initialize internal variable
 - until the end of input YUV file, call encoding function in TEncTop class
 - delete allocated buffers
 - destroy internal class
 .
 */

#ifdef HEVC_SVC
Void TAppEncTop::encodeInit(Int layer)  // initialize encoder for multiple layers
{
  // initialize internal class & member variables
  xInitLibCfg(layer);   // pass coding parameters to TEncCfg/TEncTop
  xCreateLib(layer);
  xInitLib(layer);
  
  for(int i=0; i<MAX_LAYERS; i++)
   m_cTEncTop[layer].setLayerEncoder(i, m_cTEncTop+i);
}

Void TAppEncTop::upSampleInit(Int layer)  // up sample init  // ggggg
{
    if (layer == 1)
    {
      //XXIU: in cfp, the sizes of enhance layer always do not need to be padded, so maiPad[0] and maiPad[1] are always 0;
      //for base layer, when 2x downscale from 1920x1080 to 960x540, the base layer needs to be padded, so the padding should be considered.
      int ww = m_cTEncTop[layer-1].getSourceWidth()  - m_cTEncTop[layer-1].getPad(0); //ESRYU
      int hh = m_cTEncTop[layer-1].getSourceHeight() - m_cTEncTop[layer-1].getPad(1); //ESRYU
      upSampleRatio_e upsRatio;

      int wwUp = m_cTEncTop[layer].getSourceWidth()  - m_cTEncTop[layer].getPad(0); //ESRYU;
      int hhUp = m_cTEncTop[layer].getSourceHeight() - m_cTEncTop[layer].getPad(1); //ESRYU;
      if (m_cIlpPic[0] == NULL)
      {
        for (int jj=0;jj<1;jj++)
        {
          m_cIlpPic[jj] = new  TComPic;
          m_cIlpPic[jj]->create(wwUp, hhUp, g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth, true);  //XXIU

          int numCUs = m_cIlpPic[jj]->getPicSym()->getNumberOfCUsInFrame();
          for (int xx=0;xx<numCUs;xx++)
          {
             m_cIlpPic[jj]->getPicSym()->getCU(xx)->initCU(m_cIlpPic[jj],xx);
          }
        }
      }
      m_upSampleFilter[layer].setPicutreWidthForUpsample(ww);
      m_upSampleFilter[layer].setPicutreHeightForUpsample(hh);
      if ((wwUp == ww)&& (hhUp = hh))
      {
        upsRatio = ONE_TO_ONE;
      }    
      else if ((wwUp == ww<<1) && (hhUp == hh<<1))
      {
        upsRatio = ONE_TO_TWO;
      }
      else  if ((2*wwUp == 3*ww) && (2*hhUp == 3*hh))
      {
        upsRatio = TWO_TO_THREE;
      }
      else
      {
        printf("Illegal up sample ratio WW %d to %d HH  %d to %d\n",ww,wwUp,hh,hhUp);
        assert(0);
      }
      m_upSampleFilter[layer].setUpSampleRatio(upsRatio);
    }
}
#endif

#ifdef HEVC_SVC
Void TAppEncTop::encode( Int maxLayerNum, std::fstream *outputBitstream )
{
  Int i;

  TComPicYuv*       pcPicYuvOrg[MAX_LAYERS] ;
  TComPicYuv*       pcPicYuvRec[MAX_LAYERS] ;

  // main encoder loop
  Int   iNumEncoded[MAX_LAYERS];
  Bool  bEos = false;
  Bool  bEosFirst = true, bSeqFinished = false;  //XXIU
  unsigned uiFrames = 0;
  
  list<AccessUnit> outputAccessUnits[MAX_LAYERS]; ///< list of access units to write out.  is populated by the encoding process

  for (i = 0; i < maxLayerNum; i ++)
  {
    pcPicYuvRec[i] = NULL;
    pcPicYuvOrg[i] = new TComPicYuv;
    iNumEncoded[i] = 0;
  }

  // allocate original YUV buffer
  for (i = 0; i < maxLayerNum; i ++)
  {
    pcPicYuvOrg[i]->create( m_cTEncTop[i].getSourceWidth(), m_cTEncTop[i].getSourceHeight(), 
                       m_cTEncTop[i].getMaxCUWidth(), m_cTEncTop[i].getMaxCUHeight(), m_cTEncTop[i].getMaxCUDepth());
  }  

  bool  is_GOP_Ready= false;
  int iNumPicRcvdThisGop = 0;

  for (uiFrames = 0; uiFrames < m_iFrameToBeEncoded; uiFrames ++)  //XXIU bug fix
  {
    for (i = 0; i < maxLayerNum; i ++)
    {
      UInt uiSkip = ( 1 << m_cTEncTop[i].getTemporalResolution() );
           
      is_GOP_Ready = false;

      if (uiFrames % uiSkip == 0)
      {
        // get buffers
        xGetBuffer(pcPicYuvRec[i], i);
         
        // read input YUV file
        m_cTVideoIOYuvInputFile[i].read( pcPicYuvOrg[i], m_cTEncTop[i].getPad() );

        bEos = ( m_cTVideoIOYuvInputFile[i].isEof() == 1 ?   true : false  );
        if (bEos) 
          break;

        // increase number of received frames
        m_iFrameRcvd[i] ++;

        if (m_iFrameRcvd[i] <= m_iFrameToBeEncoded)
        {
          is_GOP_Ready = m_cTEncTop[i].encode( bEos, pcPicYuvOrg[i], m_cListPicYuvRec[i], outputAccessUnits[i], iNumEncoded[i], i );
          if (i==0)  iNumPicRcvdThisGop++;
          if(m_iFrameRcvd[i] == m_iFrameToBeEncoded)  
            is_GOP_Ready = 1;  //XXIU bug fix          
        }                
      }
    }

    if(bEos)  //XXIU bug fix
    {
      if(bEosFirst)  
        is_GOP_Ready = 1;
      else
      {
        bEosFirst = false;
        break;
      }
    }

    UInt m_iHrchDepth = this->getMaxTemporalLayers();
    if (!uiFrames) m_iHrchDepth = 1;    // exception for the first frame

    if(is_GOP_Ready == true && !bSeqFinished)  //XXIU bug fix
    {
      //test if seq is finished
      if(m_iFrameRcvd[0] == m_iFrameToBeEncoded || (bEos && bEosFirst))  //XXIU
        bSeqFinished = true;
      
      int gopSize = m_iGOPSize;  // could also use m_cTEncTop[0].getGOPSize()

      if(m_cTEncTop[0].getLastPoc() == 0)  // first GOP is a single picture
      {
        gopSize = 1;
      }
    
      for ( Int iGOPid=0; iGOPid < gopSize; iGOPid++ )  
      {
        Int iTimeOffset = m_cTEncTop[0].getGOPEntry(iGOPid).m_POC;
        if(m_cTEncTop[0].getLastPoc() == 0)
        {
          iTimeOffset = 1;
        }

        if (iTimeOffset <= iNumPicRcvdThisGop)
        {
          TComList<TComPic*>*listOfPicsBL = m_cTEncTop[0].getListPic();  // SVC_NONCOLL
          for (i = 0; i < maxLayerNum; i ++) 
            m_cTEncTop[i].encodeGOP(m_cListPicYuvRec[i], outputAccessUnits[i], i, iGOPid,iTimeOffset, m_cIlpPic, listOfPicsBL);  // SVC_NONCOLL                 
        }
      }
      iNumPicRcvdThisGop = 0;

      for (i = 0; i < maxLayerNum; i ++)
      {
        m_cTEncTop[i].encodeClean( iNumEncoded[i]);
      }

      xWriteOutput(*outputBitstream, iNumEncoded, outputAccessUnits, maxLayerNum);

      for (i = 0; i < maxLayerNum; i ++)
      {
        outputAccessUnits[i].clear();
      }
    }// GOP Ready
  }// uiFrames

  double totalBitRateA = 0.0;
  double totalBitRateI = 0.0;
  double totalBitRateP = 0.0;
  double totalBitRateB = 0.0;
  double AA=0.0, II=0.0, PP=0.0, BB=0.0;
  for (int layer=0;layer<maxLayerNum;layer++)
  {
    m_cTEncTop[layer].printOutSummary(layer);
    m_cTEncTop[layer].getBitRates(layer,&AA,&II,&PP,&BB);        
    totalBitRateA = totalBitRateA + AA;
    totalBitRateI = totalBitRateI + II;
    totalBitRateP = totalBitRateP + PP;
    totalBitRateB = totalBitRateB + BB;
  }
  printf("\nTotal Bit rate summary --------------------------------------------------------\n");
  printf("Total Bit rate All %f\n",totalBitRateA);
  printf("Total Bit rate I %f\n",totalBitRateI);
  printf("Total Bit rate P %f\n",totalBitRateP);
  printf("Total Bit rate B %f\n",totalBitRateB);
  printf("\n\n");

  // delete original YUV buffer
  for ( i = maxLayerNum - 1; i >= 0; i --)
  {
    pcPicYuvOrg[i]->destroy();
    delete pcPicYuvOrg[i];
    pcPicYuvOrg[i] = NULL;
    // delete used buffers in encoder class
    m_cTEncTop[i].deletePicBuffer();
  
   // delete buffers & classes
    xDeleteBuffer(i);
    xDestroyLib(i); 
  }
#ifdef HEVC_SVC

  if (m_cIlpPic[0]!=NULL)
  {
    for (int xx=0;xx<1;xx++)
    {
      m_cIlpPic[xx]->destroy();
      delete m_cIlpPic[xx];
      m_cIlpPic[xx] = NULL;
    }
  }

#endif

  destroyROM();
  int totalBytes = 0;
  for (i = 0; i < maxLayerNum; i ++)
    totalBytes = totalBytes + printRateSummary(i);

  printf("Total bytes written to file all layers %d\n",totalBytes);
  return;
}
#else
Void TAppEncTop::encode()
{
  fstream bitstreamFile(m_pchBitstreamFile, fstream::binary | fstream::out);
  if (!bitstreamFile)
  {
    fprintf(stderr, "\nfailed to open bitstream file `%s' for writing\n", m_pchBitstreamFile);
    exit(EXIT_FAILURE);
  }

  TComPicYuv*       pcPicYuvOrg = new TComPicYuv;
  TComPicYuv*       pcPicYuvRec = NULL;
  
  // initialize internal class & member variables
  xInitLibCfg();
  xCreateLib();
  xInitLib();
  
  // main encoder loop
  Int   iNumEncoded = 0;
  Bool  bEos = false;
  
  list<AccessUnit> outputAccessUnits; ///< list of access units to write out.  is populated by the encoding process

  // allocate original YUV buffer
  pcPicYuvOrg->create( m_iSourceWidth, m_iSourceHeight, m_uiMaxCUWidth, m_uiMaxCUHeight, m_uiMaxCUDepth );
  
  while ( !bEos )
  {
    // get buffers
    xGetBuffer(pcPicYuvRec);

    // read input YUV file
    m_cTVideoIOYuvInputFile.read( pcPicYuvOrg, m_aiPad );
    
    // increase number of received frames
    m_iFrameRcvd++;
    
    // check end of file
    bEos = ( m_cTVideoIOYuvInputFile.isEof() == 1 ?   true : false  );
    bEos = ( m_iFrameRcvd == m_iFrameToBeEncoded ?    true : bEos   );
    
    // call encoding function for one frame
    m_cTEncTop.encode( bEos, pcPicYuvOrg, m_cListPicYuvRec, outputAccessUnits, iNumEncoded );
    
    // write bistream to file if necessary
    if ( iNumEncoded > 0 )
    {
      xWriteOutput(bitstreamFile, iNumEncoded, outputAccessUnits);
      outputAccessUnits.clear();
    }
  }
  // delete original YUV buffer
  pcPicYuvOrg->destroy();
  delete pcPicYuvOrg;
  pcPicYuvOrg = NULL;
  
  // delete used buffers in encoder class
  m_cTEncTop.deletePicBuffer();
  
  // delete buffers & classes
  xDeleteBuffer();
  xDestroyLib();
  
  printRateSummary();

  return;
}
#endif

// ====================================================================================================================
// Protected member functions
// ====================================================================================================================

/**
 - application has picture buffer list with size of GOP
 - picture buffer list acts as ring buffer
 - end of the list has the latest picture
 .
 */

#ifdef HEVC_SVC
Void TAppEncTop::xGetBuffer( TComPicYuv*& rpcPicYuvRec, Int layer)
{
  assert( m_iGOPSize > 0 );
  
  // org. buffer
  if ( m_cListPicYuvRec[layer].size() == (UInt)m_iGOPSize )
  {
    rpcPicYuvRec = m_cListPicYuvRec[layer].popFront();
  }
  else
  {
    rpcPicYuvRec = new TComPicYuv;
    
    rpcPicYuvRec->create( m_cTEncTop[layer].getSourceWidth(), m_cTEncTop[layer].getSourceHeight(), 
                          m_cTEncTop[layer].getMaxCUWidth(), m_cTEncTop[layer].getMaxCUHeight(), m_cTEncTop[layer].getMaxCUDepth());
  }
  m_cListPicYuvRec[layer].pushBack( rpcPicYuvRec );
}

Void TAppEncTop::xDeleteBuffer( Int layer )
{
  TComList<TComPicYuv*>::iterator iterPicYuvRec  = m_cListPicYuvRec[layer].begin();
  
  Int iSize = Int( m_cListPicYuvRec[layer].size() );
  
  for ( Int i = 0; i < iSize; i++ )
  {
    TComPicYuv*  pcPicYuvRec  = *(iterPicYuvRec++);
    pcPicYuvRec->destroy();
    delete pcPicYuvRec; pcPicYuvRec = NULL;
  }
  
}

/** \param iNumEncoded  number of encoded frames
 */
Void TAppEncTop::xWriteOutput(std::ostream& bitstreamFile, Int* iNumEncoded, const std::list<AccessUnit> *accessUnits, Int maxLayerNum)
{
    Int i, j;
    Int maxNumEncoded = 0;

    TComList<TComPicYuv*>::iterator iterPicYuvRec[MAX_LAYERS];
    list<AccessUnit>::const_iterator iterBitstream[MAX_LAYERS];

    for (i = 0; i < maxLayerNum; i ++)
    {
        iterPicYuvRec[i] = m_cListPicYuvRec[i].end();
        iterBitstream[i] = accessUnits[i].begin();
    }

    for (i = 0; i < maxLayerNum; i ++)
    {
        for (j = 0; j < iNumEncoded[i]; j ++)
        {
            --iterPicYuvRec[i];     // Rec YUV list order is 1,2,3,4,5,6,7,8
        }

        for (j = 0; j < iNumEncoded[i]; j ++)
        {
            TComPicYuv*  pcPicYuvRec  = *(iterPicYuvRec[i]++);
            if (m_cTEncTop[i].getReconFile())
                m_cTVideoIOYuvReconFile[i].write( pcPicYuvRec, m_cTEncTop[i].getCropLeft(), m_cTEncTop[i].getCropRight(), m_cTEncTop[i].getCropTop(), m_cTEncTop[i].getCropTop());
            
        }
    }

    maxNumEncoded = 0;
    for (i = 0; i < maxLayerNum; i ++)
    {
        maxNumEncoded = (iNumEncoded[i] > maxNumEncoded) ? iNumEncoded[i] : maxNumEncoded;
    }

    for ( j = 0; j < maxNumEncoded; j ++)
    {
        for ( i = 0; i < maxLayerNum; i ++)
        {
            if ( j < iNumEncoded[i] )
            {
                const AccessUnit& au = *(iterBitstream[i]++);   // access unit order is 8,4,2,6,1,3,5,7
                const vector<unsigned>& stats = writeAnnexB(bitstreamFile, au);
                rateStatsAccum(au, stats,i);
            }
        }
    }

}
#else
Void TAppEncTop::xGetBuffer( TComPicYuv*& rpcPicYuvRec)
{
  assert( m_iGOPSize > 0 );
  
  // org. buffer
  if ( m_cListPicYuvRec.size() == (UInt)m_iGOPSize )
  {
    rpcPicYuvRec = m_cListPicYuvRec.popFront();

  }
  else
  {
    rpcPicYuvRec = new TComPicYuv;
    
    rpcPicYuvRec->create( m_iSourceWidth, m_iSourceHeight, m_uiMaxCUWidth, m_uiMaxCUHeight, m_uiMaxCUDepth );

  }
  m_cListPicYuvRec.pushBack( rpcPicYuvRec );
}

Void TAppEncTop::xDeleteBuffer( )
{
  TComList<TComPicYuv*>::iterator iterPicYuvRec  = m_cListPicYuvRec.begin();
  
  Int iSize = Int( m_cListPicYuvRec.size() );
  
  for ( Int i = 0; i < iSize; i++ )
  {
    TComPicYuv*  pcPicYuvRec  = *(iterPicYuvRec++);
    pcPicYuvRec->destroy();
    delete pcPicYuvRec; pcPicYuvRec = NULL;
  }
  
}

/** \param iNumEncoded  number of encoded frames
 */
Void TAppEncTop::xWriteOutput(std::ostream& bitstreamFile, Int iNumEncoded, const std::list<AccessUnit>& accessUnits)
{
  Int i;
  
  TComList<TComPicYuv*>::iterator iterPicYuvRec = m_cListPicYuvRec.end();
  list<AccessUnit>::const_iterator iterBitstream = accessUnits.begin();
  
  for ( i = 0; i < iNumEncoded; i++ )
  {
    --iterPicYuvRec;
  }
  
  for ( i = 0; i < iNumEncoded; i++ )
  {
    TComPicYuv*  pcPicYuvRec  = *(iterPicYuvRec++);
    if (m_pchReconFile)
    {
      m_cTVideoIOYuvReconFile.write( pcPicYuvRec, m_cropLeft, m_cropRight, m_cropTop, m_cropBottom );
    }

    const AccessUnit& au = *(iterBitstream++);
    const vector<unsigned>& stats = writeAnnexB(bitstreamFile, au);
    rateStatsAccum(au, stats);
  }
}
#endif

/**
 *
 */
#ifdef HEVC_SVC
void TAppEncTop::rateStatsAccum(const AccessUnit& au, const std::vector<unsigned>& annexBsizes, Int layer)
#else
void TAppEncTop::rateStatsAccum(const AccessUnit& au, const std::vector<unsigned>& annexBsizes)
#endif
{
  AccessUnit::const_iterator it_au = au.begin();
  vector<unsigned>::const_iterator it_stats = annexBsizes.begin();

  for (; it_au != au.end(); it_au++, it_stats++)
  {
    switch ((*it_au)->m_nalUnitType)
    {
#if NAL_UNIT_TYPES_J1003_D7
    case NAL_UNIT_CODED_SLICE_TRAIL_R:
    case NAL_UNIT_CODED_SLICE_TRAIL_N:
    case NAL_UNIT_CODED_SLICE_TLA:
    case NAL_UNIT_CODED_SLICE_TSA_N:
    case NAL_UNIT_CODED_SLICE_STSA_R:
    case NAL_UNIT_CODED_SLICE_STSA_N:
    case NAL_UNIT_CODED_SLICE_BLA:
    case NAL_UNIT_CODED_SLICE_BLANT:
    case NAL_UNIT_CODED_SLICE_BLA_N_LP:
    case NAL_UNIT_CODED_SLICE_IDR:
    case NAL_UNIT_CODED_SLICE_IDR_N_LP:
    case NAL_UNIT_CODED_SLICE_CRA:
    case NAL_UNIT_CODED_SLICE_DLP:
    case NAL_UNIT_CODED_SLICE_TFD:
#else
    case NAL_UNIT_CODED_SLICE:
    case NAL_UNIT_CODED_SLICE_TFD:
    case NAL_UNIT_CODED_SLICE_TLA:
    case NAL_UNIT_CODED_SLICE_CRA:
    case NAL_UNIT_CODED_SLICE_CRANT:
    case NAL_UNIT_CODED_SLICE_BLA:
    case NAL_UNIT_CODED_SLICE_BLANT:
    case NAL_UNIT_CODED_SLICE_IDR:
#endif
    case NAL_UNIT_VPS:
    case NAL_UNIT_SPS:
    case NAL_UNIT_PPS:
#ifdef HEVC_SVC
      m_essentialBytes[layer]+= *it_stats; 
#else
      m_essentialBytes+= *it_stats; 
#endif
      break;
    default:
      break;
    }

#ifdef HEVC_SVC
    m_totalBytes[layer]+= *it_stats;
#else
    m_totalBytes+= *it_stats; 
#endif
  }
}

#ifdef HEVC_SVC
Int TAppEncTop::printRateSummary( Int layer)
{
  double time = (double) m_iFrameRcvd[layer] / m_iFrameRate;
  printf("Layer %d Bytes written to file: %u (%.3f kbps)\n",layer, m_totalBytes[layer], 0.008 * m_totalBytes[layer] / time);
#if VERBOSE_RATE
  printf("Bytes for SPS/PPS/Slice (Incl. Annex B): %u (%.3f kbps)\n", m_essentialBytes[layer], 0.008 * m_essentialBytes[layer]/ time);
#endif
  return m_totalBytes[layer];
}
#else
void TAppEncTop::printRateSummary()
{
  double time = (double) m_iFrameRcvd / m_iFrameRate;
  printf("Bytes written to file: %u (%.3f kbps)\n", m_totalBytes, 0.008 * m_totalBytes / time);
#if VERBOSE_RATE
  printf("Bytes for SPS/PPS/Slice (Incl. Annex B): %u (%.3f kbps)\n", m_essentialBytes, 0.008 * m_essentialBytes / time);
#endif
}
#endif

#ifdef HEVC_SVC
Int TAppEncTop::getLogFactor( Double r0, Double r1 )
{
  Double dLog2Factor  = log10( r1 / r0 ) / log10( 2.0 );
  Double dRound       = floor( dLog2Factor + 0.5 );
  Double dEpsilon     = 0.0001;

  if( dLog2Factor-dEpsilon < dRound && dRound < dLog2Factor+dEpsilon )
  {
    return (Int)(dRound);
  }
  return -1;
}
#endif
//! \}
