/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TAppEncTop.h
    \brief    Encoder application class (header)
*/

#ifndef __TAPPENCTOP__
#define __TAPPENCTOP__

#include <list>
#include <ostream>

#include "TLibEncoder/TEncTop.h"
#include "TLibVideoIO/TVideoIOYuv.h"
#include "TLibCommon/AccessUnit.h"
#include "TAppEncCfg.h"

//! \ingroup TAppEncoder
//! \{

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// encoder application class
class TAppEncTop : public TAppEncCfg
{
private:
  // class interface
#ifdef HEVC_SVC
  TEncTop                    m_cTEncTop[MAX_LAYERS];                    ///< encoder class

  TComUpSampleFilter         m_upSampleFilter[MAX_LAYERS];
  //TComPic*                   m_cIlpPic[MAX_NUM_REF];                    ///<  Inter layer Prediction picture =  upsampled picture 
  TComPic*                   m_cIlpPic[1];                    ///<  yong only keep one ref pic so far 

  TVideoIOYuv                m_cTVideoIOYuvInputFile[MAX_LAYERS];       ///< input YUV file
  TVideoIOYuv                m_cTVideoIOYuvReconFile[MAX_LAYERS];       ///< output reconstruction file
  
  TComList<TComPicYuv*>      m_cListPicYuvRec[MAX_LAYERS];              ///< list of reconstruction YUV files
  
  Int                        m_iFrameRcvd[MAX_LAYERS];                  ///< number of received frames

  unsigned m_totalBytes[MAX_LAYERS];
  unsigned m_essentialBytes[MAX_LAYERS];;
#else
  TEncTop                    m_cTEncTop;                    ///< encoder class
  TVideoIOYuv                m_cTVideoIOYuvInputFile;       ///< input YUV file
  TVideoIOYuv                m_cTVideoIOYuvReconFile;       ///< output reconstruction file
  
  TComList<TComPicYuv*>      m_cListPicYuvRec;              ///< list of reconstruction YUV files
  
  Int                        m_iFrameRcvd;                  ///< number of received frames
  
  unsigned m_essentialBytes;
  unsigned m_totalBytes;
#endif

protected:
  // initialization
#ifdef HEVC_SVC
  Void  xCreateLib        ( Int layer );                               ///< create files & encoder class
  Void  xInitLibCfg       ( Int layer );                               ///< initialize internal variables
  Void  xInitLib          ( Int layer );                               ///< initialize encoder class
  Void  xDestroyLib       ( Int layer );                               ///< destroy encoder class

  /// obtain required buffers
  Void xGetBuffer(TComPicYuv*& rpcPicYuvRec,  Int layer );
  
  /// delete allocated buffers
  Void  xDeleteBuffer     ( Int layer );
  
  // file I/O
  Void xWriteOutput(std::ostream& bitstreamFile, Int *iNumEncoded, const std::list<AccessUnit> *accessUnits, Int maxLayerNum ); ///< write bitstream to file
  Void rateStatsAccum(const AccessUnit& au, const std::vector<unsigned>& stats, Int layer);
  Int printRateSummary( Int layer);
#else
  Void  xCreateLib        ();                               ///< create files & encoder class
  Void  xInitLibCfg       ();                               ///< initialize internal variables
  Void  xInitLib          ();                               ///< initialize encoder class
  Void  xDestroyLib       ();                               ///< destroy encoder class
  
  /// obtain required buffers
  Void xGetBuffer(TComPicYuv*& rpcPicYuvRec);
  
  /// delete allocated buffers
  Void  xDeleteBuffer     ();
  
  // file I/O
  Void xWriteOutput(std::ostream& bitstreamFile, Int iNumEncoded, const std::list<AccessUnit>& accessUnits); ///< write bitstream to file
  void rateStatsAccum(const AccessUnit& au, const std::vector<unsigned>& stats);
  void printRateSummary();
#endif
  
public:
  TAppEncTop();
  virtual ~TAppEncTop();
  
#ifdef HEVC_SVC
  Void        encodeInit  ( Int maxLayerNum );                              ///< main encoding Initilization
  Void        upSampleInit(Int layer);                                      ///< up sample init
  Void        encode      ( Int layer, std::fstream *outputBitstream );     ///< main encoding function
  Int         getLogFactor( Double r0, Double r1 );
#else
  Void        encode      ();                               ///< main encoding function
  TEncTop&    getTEncTop  ()   { return  m_cTEncTop; }      ///< return encoder class pointer reference
#endif

};// END CLASS DEFINITION TAppEncTop

//! \}

#endif // __TAPPENCTOP__

